@@ -26,6 +26,7 @@
 
 #include <stdio.h>
 #include <unistd.h>
+#include <inttypes.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <string.h>
@@ -42,6 +43,8 @@
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/core/signal.h>
+#include <osmocom/core/endian.h>
+#include <osmocom/gsm/i460_mux.h>
 #include <osmocom/abis/e1_input.h>
 
 #define NUM_E1_TS	32
@@ -115,12 +118,19 @@ struct fake_linux_lapd_header {
 } __attribute__((packed));
 
 struct lapd_header {
+#if OSMO_IS_LITTLE_ENDIAN
 	uint8_t ea1 : 1;
 	uint8_t cr : 1;
 	uint8_t sapi : 6;
 	uint8_t ea2 : 1;
 	uint8_t tei : 7;
 	uint8_t control_foo; /* fake UM's ... */
+#elif OSMO_IS_BIG_ENDIAN
+/* auto-generated from the little endian part above (libosmocore/contrib/struct_endianess.py) */
+	uint8_t sapi:6, cr:1, ea1:1;
+	uint8_t tei:7, ea2:1;
+	uint8_t control_foo;
+#endif
 } __attribute__((packed));
 
 osmo_static_assert(offsetof(struct fake_linux_lapd_header, hatype) == 2,    hatype_offset);
@@ -134,7 +144,7 @@ static int pcap_fd = -1;
 
 int e1_set_pcap_fd(int fd)
 {
-	struct pcap_hdr header = {
+	const struct pcap_hdr header = {
 		.magic_number	= 0xa1b2c3d4,
 		.version_major	= 2,
 		.version_minor	= 4,
@@ -143,9 +153,34 @@ int e1_set_pcap_fd(int fd)
 		.snaplen	= 65535,
 		.network	= DLT_LINUX_LAPD,
 	};
+	struct e1inp_line *line;
+	int i;
 
+	/* write header */
+	if (fd >= 0) {
+		int rc = write(fd, &header, sizeof(header));
+		if (rc < 0)
+			return rc;
+	}
+
+	/* update fd in all lines in our global list of e1 lines */
+	llist_for_each_entry(line, &e1inp_line_list, list) {
+		/* Set the PCAP file descriptor for all timeslots that have
+		 * software LAPD instances, to ensure the osmo_lapd_pcap code is
+		 * used to write PCAP files (if requested) */
+		for (i = 0; i < ARRAY_SIZE(line->ts); i++) {
+			struct e1inp_ts *e1i_ts = &line->ts[i];
+			if (e1i_ts->lapd)
+				e1i_ts->lapd->pcap_fd = fd;
+		}
+	}
+
+	/* close previous and update global */
+	if (pcap_fd >= 0)
+		close(pcap_fd);
 	pcap_fd = fd;
-	return write(pcap_fd, &header, sizeof(header));
+
+	return 0;
 }
 
 /* This currently only works for the D-Channel */
@@ -207,12 +242,13 @@ const char *e1inp_signtype_name(enum e1inp_sign_type tp)
 	return get_value_string(e1inp_sign_type_names, tp);
 }
 
-const struct value_string e1inp_ts_type_names[6] = {
+const struct value_string e1inp_ts_type_names[] = {
 	{ E1INP_TS_TYPE_NONE,	"None" },
 	{ E1INP_TS_TYPE_SIGN,	"Signalling" },
 	{ E1INP_TS_TYPE_TRAU,	"TRAU" },
 	{ E1INP_TS_TYPE_RAW,	"RAW" },
 	{ E1INP_TS_TYPE_HDLC,	"HDLC" },
+	{ E1INP_TS_TYPE_I460,	"I460" },
 	{ 0, NULL }
 };
 
@@ -264,7 +300,7 @@ int abis_rsl_sendmsg(struct msgb *msg)
 /* Timeslot */
 int e1inp_ts_config_trau(struct e1inp_ts *ts, struct e1inp_line *line,
 			 int (*trau_rcv_cb)(struct subch_demux *dmx, int ch,
-					uint8_t *data, int len, void *_priv))
+					const ubit_t *data, int len, void *_priv))
 {
 	if (ts->type == E1INP_TS_TYPE_TRAU && ts->line && line)
 		return 0;
@@ -279,6 +315,25 @@ int e1inp_ts_config_trau(struct e1inp_ts *ts, struct e1inp_line *line,
 	return 0;
 }
 
+int e1inp_ts_config_i460(struct e1inp_ts *ts, struct e1inp_line *line)
+{
+	if (ts->type == E1INP_TS_TYPE_I460 && ts->line && line)
+		return 0;
+
+	ts->type = E1INP_TS_TYPE_I460;
+	ts->line = line;
+	osmo_i460_ts_init(&ts->i460.i460_ts);
+	return 0;
+}
+
+void e1inp_ts_name(char *out, size_t out_len, const struct e1inp_ts *ts)
+{
+	if (ts->line->name)
+		snprintf(out, out_len, "%s:%u", ts->line->name, ts->num);
+	else
+		snprintf(out, out_len, "%u:%u", ts->line->num, ts->num);
+}
+
 int e1inp_ts_config_sign(struct e1inp_ts *ts, struct e1inp_line *line)
 {
 	if (ts->type == E1INP_TS_TYPE_SIGN && ts->line && line)
@@ -325,6 +380,48 @@ int e1inp_ts_config_hdlc(struct e1inp_ts *ts, struct e1inp_line *line,
 	return 0;
 }
 
+static int e1inp_line_use_cb(struct osmo_use_count_entry *use_count_entry, int32_t old_use_count,
+			     const char *file, int file_line)
+{
+	char buf[512];
+	struct osmo_use_count *uc = use_count_entry->use_count;
+	struct e1inp_line *line = uc->talloc_object;
+
+	LOGPSRC(DLINP, LOGL_INFO, file, file_line,
+		"E1L(%u) Line (%p) reference count %s changed %" PRId32 " -> %" PRId32 " [%s]\n",
+		(line)->num, line, use_count_entry->use,
+		old_use_count, use_count_entry->count,
+		osmo_use_count_name_buf(buf, sizeof(buf), uc));
+
+	if (!use_count_entry->count)
+		osmo_use_count_free(use_count_entry);
+
+	if (osmo_use_count_total(uc) > 0)
+		return 0;
+
+	/* Remove our counter group from libosmocore's global counter
+	 * list if we are freeing the last remaining talloc context.
+	 * Otherwise we get a use-after-free when libosmocore's timer
+	 * ticks again and attempts to update these counters (OS#3011).
+	 *
+	 * Note that talloc internally counts "secondary" references
+	 * _in addition to_ the initial allocation context, so yes,
+	 * we must check for *zero* remaining secondary contexts here. */
+	if (talloc_reference_count(line->rate_ctr) == 0) {
+		rate_ctr_group_free(line->rate_ctr);
+	} else {
+		/* We are not freeing the last talloc context.
+		 * Instead of calling talloc_free(), unlink this 'line' pointer
+		 * which serves as one of several talloc contexts for the rate
+		 * counters and driver private state. */
+		talloc_unlink(line, line->rate_ctr);
+		if (line->driver_data)
+			talloc_unlink(line, line->driver_data);
+	}
+	talloc_free(line);
+	return 0;
+}
+
 struct e1inp_line *e1inp_line_find(uint8_t e1_nr)
 {
 	struct e1inp_line *e1i_line;
@@ -346,8 +443,7 @@ e1inp_line_create(uint8_t e1_nr, const char *driver_name)
 
 	line = e1inp_line_find(e1_nr);
 	if (line) {
-		LOGP(DLINP, LOGL_ERROR, "E1 Line %u already exists\n",
-		     e1_nr);
+		LOGPIL(line, DLINP, LOGL_ERROR, "E1 Line %u already exists\n", e1_nr);
 		return NULL;
 	}
 
@@ -367,7 +463,7 @@ e1inp_line_create(uint8_t e1_nr, const char *driver_name)
 
 	line->rate_ctr = rate_ctr_group_alloc(line, &e1inp_ctr_g_d, line->num);
 	if (!line->rate_ctr) {
-		LOGP(DLINP, LOGL_ERROR, "Cannot allocate counter group\n");
+		LOGPIL(line, DLINP, LOGL_ERROR, "Cannot allocate counter group\n");
 		talloc_free(line);
 		return NULL;
 	}
@@ -377,14 +473,18 @@ e1inp_line_create(uint8_t e1_nr, const char *driver_name)
 		line->ts[i].num = i+1;
 		line->ts[i].line = line;
 	}
-	line->refcnt++;
+
+	line->use_count.talloc_object = line;
+	line->use_count.use_cb = e1inp_line_use_cb;
+	e1inp_line_get2(line, "ctor");
+
 	llist_add_tail(&line->list, &e1inp_line_list);
 
 	return line;
 }
 
 struct e1inp_line *
-e1inp_line_clone(void *ctx, struct e1inp_line *line)
+e1inp_line_clone(void *ctx, struct e1inp_line *line, const char *use)
 {
 	struct e1inp_line *clone;
 
@@ -413,47 +513,23 @@ e1inp_line_clone(void *ctx, struct e1inp_line *line)
 	if (line->driver_data)
 		clone->driver_data = talloc_reference(clone, line->driver_data);
 
-	clone->refcnt = 1;
+	clone->use_count = (struct osmo_use_count) {
+		.talloc_object = clone,
+		.use_cb = e1inp_line_use_cb,
+		.use_counts = {0},
+	};
+	e1inp_line_get2(clone, use); /* Clone is used internally for bfd */
 	return clone;
 }
 
 void e1inp_line_get(struct e1inp_line *line)
 {
-	int old_refcnt = line->refcnt++;
-
-	LOGP(DLINP, LOGL_DEBUG, "Line '%s' (%p) reference count get: %d -> %d\n",
-	     line->name, line, old_refcnt, line->refcnt);
+	e1inp_line_get2(line, "unknown");
 }
 
 void e1inp_line_put(struct e1inp_line *line)
 {
-	int old_refcnt = line->refcnt--;
-
-	LOGP(DLINP, LOGL_DEBUG, "Line '%s' (%p) reference count put: %d -> %d\n",
-	     line->name, line, old_refcnt, line->refcnt);
-
-	if (line->refcnt == 0) {
-		/* Remove our counter group from libosmocore's global counter
-		 * list if we are freeing the last remaining talloc context.
-		 * Otherwise we get a use-after-free when libosmocore's timer
-		 * ticks again and attempts to update these counters (OS#3011).
-		 *
-		 * Note that talloc internally counts "secondary" references
-		 * _in addition to_ the initial allocation context, so yes,
-		 * we must check for *zero* remaining secondary contexts here. */
-		if (talloc_reference_count(line->rate_ctr) == 0) {
-			rate_ctr_group_free(line->rate_ctr);
-		} else {
-			/* We are not freeing the last talloc context.
-			 * Instead of calling talloc_free(), unlink this 'line' pointer
-			 * which serves as one of several talloc contexts for the rate
-			 * counters and driver private state. */
-			talloc_unlink(line, line->rate_ctr);
-			if (line->driver_data)
-				talloc_unlink(line, line->driver_data);
-		}
-		talloc_free(line);
-	}
+	e1inp_line_put2(line, "unknown");
 }
 
 void
@@ -546,6 +622,8 @@ e1inp_sign_link_create(struct e1inp_ts *ts, enum e1inp_sign_type type,
 	link->tei = tei;
 	link->sapi = sapi;
 
+	e1inp_line_get2(link->ts->line, "e1inp_sign_link");
+
 	llist_add_tail(&link->list, &ts->sign.sign_links);
 
 	return link;
@@ -567,7 +645,7 @@ void e1inp_sign_link_destroy(struct e1inp_sign_link *link)
 	if (link->ts->line->driver->close)
 		link->ts->line->driver->close(link);
 
-	e1inp_line_put(link->ts->line);
+	e1inp_line_put2(link->ts->line, "e1inp_sign_link");
 	talloc_free(link);
 }
 
@@ -590,13 +668,13 @@ int e1inp_rx_ts(struct e1inp_ts *ts, struct msgb *msg,
 		/* consult the list of signalling links */
 		link = e1inp_lookup_sign_link(ts, tei, sapi);
 		if (!link) {
-			LOGP(DLMI, LOGL_ERROR, "didn't find signalling link for "
+			LOGPITS(ts, DLMI, LOGL_ERROR, "didn't find signalling link for "
 				"tei %d, sapi %d\n", tei, sapi);
 			msgb_free(msg);
 			return -EINVAL;
 		}
 		if (!ts->line->ops->sign_link) {
-	                LOGP(DLINP, LOGL_ERROR, "Fix your application, "
+	                LOGPITS(ts, DLINP, LOGL_ERROR, "Fix your application, "
 				"no action set for signalling messages.\n");
 			msgb_free(msg);
 			return -ENOENT;
@@ -614,9 +692,13 @@ int e1inp_rx_ts(struct e1inp_ts *ts, struct msgb *msg,
 	case E1INP_TS_TYPE_HDLC:
 		ts->hdlc.recv_cb(ts, msg);
 		break;
+	case E1INP_TS_TYPE_I460:
+		osmo_i460_demux_in(&ts->i460.i460_ts, msg->l2h, msgb_l2len(msg));
+		msgb_free(msg);
+		break;
 	default:
 		ret = -EINVAL;
-		LOGP(DLMI, LOGL_ERROR, "unknown TS type %u\n", ts->type);
+		LOGPITS(ts, DLMI, LOGL_ERROR, "unknown TS type %u\n", ts->type);
 		msgb_free(msg);
 		break;
 	}
@@ -644,7 +726,7 @@ int e1inp_rx_ts_lapd(struct e1inp_ts *e1i_ts, struct msgb *msg)
 	else
 		tei = msg->data[1] >> 1;
 
-	DEBUGP(DLMI, "<= len = %d, sapi(%d) tei(%d)\n", msg->len, sapi, tei);
+	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "<= len = %d, sapi(%d) tei(%d)\n", msg->len, sapi, tei);
 
 	ret = lapd_receive(e1i_ts->lapd, msg, &error);
 	if (ret < 0) {
@@ -671,18 +753,18 @@ void e1inp_dlsap_up(struct osmo_dlsap_prim *dp, uint8_t tei, uint8_t sapi,
 
 	switch (dp->oph.primitive) {
 	case PRIM_DL_EST:
-		DEBUGP(DLMI, "DL_EST: sapi(%d) tei(%d)\n", sapi, tei);
+		LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "DL_EST: sapi(%d) tei(%d)\n", sapi, tei);
 		e1inp_event(e1i_ts, S_L_INP_TEI_UP, tei, sapi);
 		break;
 	case PRIM_DL_REL:
-		DEBUGP(DLMI, "DL_REL: sapi(%d) tei(%d)\n", sapi, tei);
+		LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "DL_REL: sapi(%d) tei(%d)\n", sapi, tei);
 		e1inp_event(e1i_ts, S_L_INP_TEI_DN, tei, sapi);
 		break;
 	case PRIM_DL_DATA:
 	case PRIM_DL_UNIT_DATA:
 		if (dp->oph.operation == PRIM_OP_INDICATION) {
 			msg->l2h = msg->l3h;
-			DEBUGP(DLMI, "RX: %s sapi=%d tei=%d\n",
+			LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "RX: %s sapi=%d tei=%d\n",
 				osmo_hexdump(msgb_l2(msg), msgb_l2len(msg)),
 				sapi, tei);
 			e1inp_rx_ts(e1i_ts, msg, tei, sapi);
@@ -690,7 +772,7 @@ void e1inp_dlsap_up(struct osmo_dlsap_prim *dp, uint8_t tei, uint8_t sapi,
 		}
 		break;
 	case PRIM_MDL_ERROR:
-		DEBUGP(DLMI, "MDL_EERROR: cause(%d)\n", dp->u.error_ind.cause);
+		LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "MDL_EERROR: cause(%d)\n", dp->u.error_ind.cause);
 		break;
 	default:
 		printf("ERROR: unknown prim\n");
@@ -730,8 +812,7 @@ struct msgb *e1inp_tx_ts(struct e1inp_ts *e1i_ts,
 			return NULL;
 		len = subchan_mux_out(&e1i_ts->trau.mux, msg->data, 40);
 		if (len != 40) {
-			LOGP(DLMI, LOGL_ERROR,
-			     "cannot transmit, failed to mux\n");
+			LOGPITS(e1i_ts, DLMI, LOGL_ERROR, "cannot transmit, failed to mux\n");
 			msgb_free(msg);
 			return NULL;
 		}
@@ -745,15 +826,21 @@ struct msgb *e1inp_tx_ts(struct e1inp_ts *e1i_ts,
 		/* Get msgb from tx_queue */
 		msg = msgb_dequeue(&e1i_ts->hdlc.tx_queue);
 		break;
+	case E1INP_TS_TYPE_I460:
+		msg = msgb_alloc(TSX_ALLOC_SIZE, "I460_TX");
+		if (!msg)
+			return NULL;
+		len = osmo_i460_mux_out(&e1i_ts->i460.i460_ts, msg->data, 160);
+		msgb_put(msg, len);
+		break;
 	default:
-		LOGP(DLMI, LOGL_ERROR, "unsupported E1 TS type %u\n", e1i_ts->type);
+		LOGPITS(e1i_ts, DLMI, LOGL_ERROR, "unsupported E1 TS type %u\n", e1i_ts->type);
 		return NULL;
 	}
 	return msg;
 }
 
-static int e1inp_int_snd_event(struct e1inp_ts *ts,
-			       struct e1inp_sign_link *link, int evt)
+int e1inp_int_snd_event(struct e1inp_ts *ts, struct e1inp_sign_link *link, int evt)
 {
 	struct input_signal_data isd;
 	isd.line = ts->line;
@@ -780,19 +867,6 @@ int e1inp_event(struct e1inp_ts *ts, int evt, uint8_t tei, uint8_t sapi)
 	return e1inp_int_snd_event(ts, link, evt);
 }
 
-void e1inp_close_socket(struct e1inp_ts *ts,
-			struct e1inp_sign_link *sign_link,
-			struct osmo_fd *bfd)
-{
-	e1inp_int_snd_event(ts, sign_link, S_L_INP_TEI_DN);
-	/* the first e1inp_sign_link_destroy call closes the socket. */
-	if (bfd->fd != -1) {
-		osmo_fd_unregister(bfd);
-		close(bfd->fd);
-		bfd->fd = -1;
-	}
-}
-
 /* register a driver with the E1 core */
 int e1inp_driver_register(struct e1inp_driver *drv)
 {
@@ -816,8 +890,6 @@ int e1inp_line_update(struct e1inp_line *line)
 	struct input_signal_data isd;
 	int i, rc;
 
-	e1inp_line_get(line);
-
 	if (line->driver && line->ops && line->driver->line_update) {
 		rc = line->driver->line_update(line);
 	} else
@@ -869,6 +941,7 @@ const struct value_string e1inp_signal_names[] = {
 
 void e1inp_misdn_init(void);
 void e1inp_dahdi_init(void);
+void e1inp_e1d_init(void);
 void e1inp_ipaccess_init(void);
 void e1inp_rs232_init(void);
 void e1inp_unixsocket_init(void);
@@ -883,6 +956,9 @@ void e1inp_init(void)
 	e1inp_misdn_init();
 #ifdef HAVE_DAHDI_USER_H
 	e1inp_dahdi_init();
+#endif
+#ifdef HAVE_E1D
+	e1inp_e1d_init();
 #endif
 	e1inp_ipaccess_init();
 	e1inp_rs232_init();
