@@ -59,23 +59,60 @@ static void *tall_ipa_ctx;
 #define DEFAULT_TCP_KEEPALIVE_INTERVAL     3
 #define DEFAULT_TCP_KEEPALIVE_RETRY_COUNT  10
 
+static inline struct e1inp_ts *ipaccess_line_ts(struct osmo_fd *bfd, struct e1inp_line *line)
+{
+	if (bfd->priv_nr == E1INP_SIGN_OML)
+		return e1inp_line_ipa_oml_ts(line);
+	else
+		return e1inp_line_ipa_rsl_ts(line, bfd->priv_nr - E1INP_SIGN_RSL);
+}
+
+static inline void ipaccess_keepalive_fsm_cleanup(struct e1inp_ts *e1i_ts)
+{
+	struct osmo_fsm_inst *ka_fsm;
+
+	ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
+	if (ka_fsm) {
+		ipa_keepalive_fsm_stop(ka_fsm);
+		e1i_ts->driver.ipaccess.ka_fsm = NULL;
+	}
+}
+
 static int ipaccess_drop(struct osmo_fd *bfd, struct e1inp_line *line)
 {
 	int ret = 1;
-	struct e1inp_ts *e1i_ts;
-	if (bfd->priv_nr == E1INP_SIGN_OML)
-		e1i_ts = e1inp_line_ipa_oml_ts(line);
-	else
-		e1i_ts = e1inp_line_ipa_rsl_ts(line, bfd->priv_nr - E1INP_SIGN_RSL);
+	struct e1inp_ts *e1i_ts = ipaccess_line_ts(bfd, line);
+	e1inp_line_get2(line, __func__);
+
+	ipaccess_keepalive_fsm_cleanup(e1i_ts);
 
 	/* Error case: we did not see any ID_RESP yet for this socket. */
 	if (bfd->fd != -1) {
-		LOGP(DLINP, LOGL_ERROR, "Forcing socket shutdown with "
-					"no signal link set\n");
+		LOGPITS(e1i_ts, DLINP, LOGL_NOTICE, "Forcing socket shutdown\n");
 		osmo_fd_unregister(bfd);
 		close(bfd->fd);
 		bfd->fd = -1;
+		switch (line->ops->cfg.ipa.role) {
+		case E1INP_LINE_R_BSC:
+			/* This is BSC code, ipaccess_drop() is only called for
+			   accepted() sockets, hence the bfd holds a reference to
+			   e1inp_line in ->data that needs to be released */
+			OSMO_ASSERT(bfd->data == line);
+			bfd->data = NULL;
+			e1inp_line_put2(line, "ipa_bfd");
+			break;
+		case E1INP_LINE_R_BTS:
+			/* BTS code: bfd->data contains pointer to struct
+			 * ipa_client_conn. Leave it alive so it reconnects.
+			 */
+			break;
+		default:
+			break;
+		}
 		ret = -ENOENT;
+	} else {
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR,
+			"Forcing socket shutdown with no signal link set\n");
 	}
 
 	msgb_free(e1i_ts->pending_msg);
@@ -84,9 +121,84 @@ static int ipaccess_drop(struct osmo_fd *bfd, struct e1inp_line *line)
 	/* e1inp_sign_link_destroy releases the socket descriptors for us. */
 	line->ops->sign_link_down(line);
 
+	e1inp_line_put2(line, __func__);
 	return ret;
 }
 
+static void ipa_bsc_keepalive_write_server_cb(struct osmo_fsm_inst *fi, void *conn, struct msgb *msg)
+{
+	struct osmo_fd *bfd = (struct osmo_fd *)conn;
+	write(bfd->fd, msg->data, msg->len);
+	msgb_free(msg);
+}
+
+static int ipa_bsc_keepalive_timeout_cb(struct osmo_fsm_inst *fi, void *data)
+{
+	struct osmo_fd *bfd = (struct osmo_fd *)data;
+
+	if (bfd->fd == -1)
+		return 1;
+
+	ipaccess_drop(bfd, (struct e1inp_line *)bfd->data);
+	return 1;
+}
+
+static void ipaccess_bsc_keepalive_fsm_alloc(struct e1inp_ts *e1i_ts, struct osmo_fd *bfd, const char *id)
+{
+	struct e1inp_line *line = e1i_ts->line;
+	struct osmo_fsm_inst *ka_fsm;
+
+	ipaccess_keepalive_fsm_cleanup(e1i_ts);
+	if (!line->ipa_kap)
+		return;
+
+	ka_fsm = ipa_generic_conn_alloc_keepalive_fsm(tall_ipa_ctx, bfd, line->ipa_kap, id);
+	e1i_ts->driver.ipaccess.ka_fsm = ka_fsm;
+	if (!ka_fsm)
+		return;
+
+	ipa_keepalive_fsm_set_timeout_cb(ka_fsm, ipa_bsc_keepalive_timeout_cb);
+	ipa_keepalive_fsm_set_send_cb(ka_fsm, ipa_bsc_keepalive_write_server_cb);
+	ipa_keepalive_fsm_start(ka_fsm);
+}
+
+static void ipa_bts_keepalive_write_client_cb(struct osmo_fsm_inst *fi, void *conn, struct msgb *msg) {
+	struct ipa_client_conn *link = (struct ipa_client_conn *)conn;
+	int ret = 0;
+
+	ret = ipa_send(link->ofd->fd, msg->data, msg->len);
+	if (ret != msg->len) {
+		LOGP(DLINP, LOGL_ERROR, "cannot send message. Reason: %s\n", strerror(errno));
+	}
+	msgb_free(msg);
+}
+
+static void update_fd_settings(struct e1inp_line *line, int fd);
+static void ipaccess_bts_updown_cb(struct ipa_client_conn *link, int up);
+
+static int ipa_bts_keepalive_timeout_cb(struct osmo_fsm_inst *fi, void *conn) {
+	ipaccess_bts_updown_cb(conn, false);
+	return 1;
+}
+
+static void ipaccess_bts_keepalive_fsm_alloc(struct e1inp_ts *e1i_ts, struct ipa_client_conn *client, const char *id)
+{
+	struct e1inp_line *line = e1i_ts->line;
+	struct osmo_fsm_inst *ka_fsm;
+
+	ipaccess_keepalive_fsm_cleanup(e1i_ts);
+	if (!line->ipa_kap)
+		return;
+
+	ka_fsm = ipa_client_conn_alloc_keepalive_fsm(client, line->ipa_kap, id);
+	e1i_ts->driver.ipaccess.ka_fsm = ka_fsm;
+	if (!ka_fsm)
+		return;
+
+	ipa_keepalive_fsm_set_timeout_cb(ka_fsm, ipa_bts_keepalive_timeout_cb);
+	ipa_keepalive_fsm_set_send_cb(ka_fsm, ipa_bts_keepalive_write_client_cb);
+}
+
 /* Returns -1 on error, and 0 or 1 on success. If -1 or 1 is returned, line has
  * been released and should not be used anymore by the caller. */
 static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
@@ -98,6 +210,14 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 	struct e1inp_sign_link *sign_link;
 	char *unitid;
 	int len, ret;
+	struct e1inp_ts *e1i_ts;
+	struct osmo_fsm_inst *ka_fsm;
+
+	/* peek the pong for our keepalive fsm */
+	e1i_ts = ipaccess_line_ts(bfd, line);
+	ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
+	if (ka_fsm && msg_type == IPAC_MSGT_PONG)
+		ipa_keepalive_fsm_pong_received(ka_fsm);
 
 	/* Handle IPA PING, PONG and ID_ACK messages. */
 	ret = ipa_ccm_rcvmsg_base(msg, bfd);
@@ -120,10 +240,10 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 
 	switch (msg_type) {
 	case IPAC_MSGT_ID_RESP:
-		DEBUGP(DLMI, "ID_RESP\n");
+		DEBUGP(DLMI, "ID_RESP ");
 		/* parse tags, search for Unit ID */
 		ret = ipa_ccm_id_resp_parse(&tlvp, (const uint8_t *)msg->l2h+1, msgb_l2len(msg)-1);
-		DEBUGP(DLMI, "\n");
+		DEBUGPC(DLMI, "\n");
 		if (ret < 0) {
 			LOGP(DLINP, LOGL_ERROR, "IPA response message "
 				"with malformed TLVs\n");
@@ -165,9 +285,12 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 					"closing socket.\n");
 				goto err;
 			}
+
+			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, bfd, "oml_bsc_to_bts");
+
 		} else if (bfd->priv_nr == E1INP_SIGN_RSL) {
 			struct e1inp_ts *ts;
-                        struct osmo_fd *newbfd;
+			struct osmo_fd *newbfd;
 			struct e1inp_line *new_line;
 
 			sign_link =
@@ -179,36 +302,45 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 					"closing socket.\n");
 				goto err;
 			}
+			/* Finally, we know which OML link is associated with
+			 * this RSL link, attach it to this socket. */
+			new_line = sign_link->ts->line;
 			/* this is a bugtrap, the BSC should be using the
 			 * virtual E1 line used by OML for this RSL link. */
-			if (sign_link->ts->line == line) {
+			if (new_line == line) {
 				LOGP(DLINP, LOGL_ERROR,
 					"Fix your BSC, you should use the "
 					"E1 line used by the OML link for "
 					"your RSL link.\n");
 				return 0;
 			}
-			/* Finally, we know which OML link is associated with
-			 * this RSL link, attach it to this socket. */
-			bfd->data = new_line = sign_link->ts->line;
-			e1inp_line_get(new_line);
+			e1inp_line_get2(new_line, "ipa_bfd");
 			ts = e1inp_line_ipa_rsl_ts(new_line, unit_data.trx_id);
 			newbfd = &ts->driver.ipaccess.fd;
+			OSMO_ASSERT(newbfd != bfd);
 
-			/* get rid of our old temporary bfd */
 			/* preserve 'newbfd->when' flags potentially set by sign_link_up() */
 			osmo_fd_setup(newbfd, bfd->fd, newbfd->when | bfd->when, bfd->cb,
-				      bfd->data, E1INP_SIGN_RSL + unit_data.trx_id);
+				      new_line, E1INP_SIGN_RSL + unit_data.trx_id);
+
+
+			/* now we can release the dummy RSL line (old temporary bfd). */
 			osmo_fd_unregister(bfd);
 			bfd->fd = -1;
+			/* bfd->data holds a reference to line, drop it */
+			OSMO_ASSERT(bfd->data == line);
+			bfd->data = NULL;
+			e1inp_line_put2(line, "ipa_bfd");
+
 			ret = osmo_fd_register(newbfd);
 			if (ret < 0) {
 				LOGP(DLINP, LOGL_ERROR,
 				     "could not register FD\n");
 				goto err;
 			}
-			/* now we can release the dummy RSL line. */
-			e1inp_line_put(line);
+
+			e1i_ts = ipaccess_line_ts(newbfd, new_line);
+			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, newbfd, "rsl_bsc_to_bts");
 			return 1;
 		}
 		break;
@@ -218,12 +350,15 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 	}
 	return 0;
 err:
-	osmo_fd_unregister(bfd);
 	if (bfd->fd != -1) {
+		osmo_fd_unregister(bfd);
 		close(bfd->fd);
 		bfd->fd = -1;
+		/* This is a BSC accepted socket, bfd->data holds a reference to line, drop it */
+		OSMO_ASSERT(bfd->data == line);
+		bfd->data = NULL;
+		e1inp_line_put2(line, "ipa_bfd");
 	}
-	e1inp_line_put(line);
 	return -1;
 }
 
@@ -238,11 +373,7 @@ static int handle_ts1_read(struct osmo_fd *bfd)
 	struct msgb *msg = NULL;
 	int ret, rc;
 
-	if (bfd->priv_nr == E1INP_SIGN_OML)
-		e1i_ts = e1inp_line_ipa_oml_ts(line);
-	else
-		e1i_ts = e1inp_line_ipa_rsl_ts(line, bfd->priv_nr - E1INP_SIGN_RSL);
-
+	e1i_ts = ipaccess_line_ts(bfd, line);
 	ret = ipa_msg_recv_buffered(bfd->fd, &msg, &e1i_ts->pending_msg);
 	if (ret < 0) {
 		if (ret == -EAGAIN)
@@ -303,7 +434,7 @@ err:
 
 static int ts_want_write(struct e1inp_ts *e1i_ts)
 {
-	e1i_ts->driver.ipaccess.fd.when |= BSC_FD_WRITE;
+	osmo_fd_write_enable(&e1i_ts->driver.ipaccess.fd);
 
 	return 0;
 }
@@ -312,7 +443,30 @@ static void ipaccess_close(struct e1inp_sign_link *sign_link)
 {
 	struct e1inp_ts *e1i_ts = sign_link->ts;
 	struct osmo_fd *bfd = &e1i_ts->driver.ipaccess.fd;
-	return e1inp_close_socket(e1i_ts, sign_link, bfd);
+	struct e1inp_line *line = e1i_ts->line;
+
+	/* line might not exist if != bsc||bts */
+	if (line) {
+		/* depending on caller the fsm might be dead */
+		struct osmo_fsm_inst* ka_fsm = ipaccess_line_ts(bfd, line)->driver.ipaccess.ka_fsm;
+		if (ka_fsm)
+			ipa_keepalive_fsm_stop(ka_fsm);
+
+	}
+
+	e1inp_int_snd_event(e1i_ts, sign_link, S_L_INP_TEI_DN);
+	/* the first e1inp_sign_link_destroy call closes the socket. */
+	if (bfd->fd != -1) {
+		osmo_fd_unregister(bfd);
+		close(bfd->fd);
+		bfd->fd = -1;
+		/* If The bfd holds a reference to e1inp_line in ->data (BSC
+		 * accepted() sockets), then release it */
+		if (bfd->data == line) {
+			bfd->data = NULL;
+			e1inp_line_put2(line, "ipa_bfd");
+		}
+	}
 }
 
 static void timeout_ts1_write(void *data)
@@ -331,12 +485,8 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 	struct msgb *msg;
 	int ret;
 
-	if (bfd->priv_nr == E1INP_SIGN_OML)
-		e1i_ts = e1inp_line_ipa_oml_ts(line);
-	else
-		e1i_ts = e1inp_line_ipa_rsl_ts(line, bfd->priv_nr - E1INP_SIGN_RSL);
-
-	bfd->when &= ~BSC_FD_WRITE;
+	e1i_ts = ipaccess_line_ts(bfd, line);
+	osmo_fd_write_disable(bfd);
 
 	/* get the next msg for this timeslot */
 	msg = e1inp_tx_ts(e1i_ts, &sign_link);
@@ -351,7 +501,7 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 	case E1INP_SIGN_OSMO:
 		break;
 	default:
-		bfd->when |= BSC_FD_WRITE; /* come back for more msg */
+		osmo_fd_write_enable(bfd); /* come back for more msg */
 		ret = -EINVAL;
 		goto out;
 	}
@@ -359,11 +509,12 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 	msg->l2h = msg->data;
 	ipa_prepend_header(msg, sign_link->tei);
 
-	DEBUGP(DLMI, "TX %u: %s\n", ts_nr, osmo_hexdump(msg->l2h, msgb_l2len(msg)));
+	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "TX %u: %s\n", ts_nr,
+		osmo_hexdump(msg->l2h, msgb_l2len(msg)));
 
 	ret = send(bfd->fd, msg->data, msg->len, 0);
 	if (ret != msg->len) {
-		LOGP(DLINP, LOGL_ERROR, "failed to send A-bis IPA signalling "
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "failed to send A-bis IPA signalling "
 			"message. Reason: %s\n", strerror(errno));
 		goto err;
 	}
@@ -402,9 +553,9 @@ int ipaccess_fd_cb(struct osmo_fd *bfd, unsigned int what)
 {
 	int rc = 0;
 
-	if (what & BSC_FD_READ)
+	if (what & OSMO_FD_READ)
 		rc = handle_ts1_read(bfd);
-	if (rc != -EBADF && (what & BSC_FD_WRITE))
+	if (rc != -EBADF && (what & OSMO_FD_WRITE))
 		rc = handle_ts1_write(bfd);
 
 	return rc;
@@ -478,6 +629,11 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 #endif
 #endif
 	}
+
+	val = 1;
+	ret = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val));
+	if (ret < 0)
+		LOGP(DLINP, LOGL_ERROR, "Failed to set TCP_NODELAY: %s\n", strerror(errno));
 }
 
 /* callback of the OML listening filedescriptor */
@@ -490,7 +646,7 @@ static int ipaccess_bsc_oml_cb(struct ipa_server_link *link, int fd)
 	struct osmo_fd *bfd;
 
 	/* clone virtual E1 line for this new OML link. */
-	line = e1inp_line_clone(tall_ipa_ctx, link->line);
+	line = e1inp_line_clone(tall_ipa_ctx, link->line, "ipa_bfd");
 	if (line == NULL) {
 		LOGP(DLINP, LOGL_ERROR, "could not clone E1 line\n");
 		return -ENOMEM;
@@ -506,7 +662,7 @@ static int ipaccess_bsc_oml_cb(struct ipa_server_link *link, int fd)
 	e1i_ts = e1inp_line_ipa_oml_ts(line);
 
 	bfd = &e1i_ts->driver.ipaccess.fd;
-	osmo_fd_setup(bfd, fd, BSC_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_OML);
+	osmo_fd_setup(bfd, fd, OSMO_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_OML);
 	ret = osmo_fd_register(bfd);
 	if (ret < 0) {
 		LOGP(DLINP, LOGL_ERROR, "could not register FD\n");
@@ -529,7 +685,8 @@ err_socket:
 err_line:
 	close(bfd->fd);
 	bfd->fd = -1;
-	e1inp_line_put(line);
+	bfd->data = NULL;
+	e1inp_line_put2(line, "ipa_bfd");
 	return ret;
 }
 
@@ -542,7 +699,7 @@ static int ipaccess_bsc_rsl_cb(struct ipa_server_link *link, int fd)
 
         /* We don't know yet which OML link to associate it with. Thus, we
          * allocate a temporary E1 line until we have received ID. */
-	line = e1inp_line_clone(tall_ipa_ctx, link->line);
+	line = e1inp_line_clone(tall_ipa_ctx, link->line, "ipa_bfd");
 	if (line == NULL) {
 		LOGP(DLINP, LOGL_ERROR, "could not clone E1 line\n");
 		return -ENOMEM;
@@ -558,7 +715,7 @@ static int ipaccess_bsc_rsl_cb(struct ipa_server_link *link, int fd)
 	e1i_ts = e1inp_line_ipa_rsl_ts(line, 0);
 
 	bfd = &e1i_ts->driver.ipaccess.fd;
-	osmo_fd_setup(bfd, fd, BSC_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_RSL);
+	osmo_fd_setup(bfd, fd, OSMO_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_RSL);
 	ret = osmo_fd_register(bfd);
 	if (ret < 0) {
 		LOGP(DLINP, LOGL_ERROR, "could not register FD\n");
@@ -579,7 +736,8 @@ err_socket:
 err_line:
 	close(bfd->fd);
 	bfd->fd = -1;
-	e1inp_line_put(line);
+	bfd->data = NULL;
+	e1inp_line_put2(line, "ipa_bfd");
 	return ret;
 }
 
@@ -684,10 +842,14 @@ static void ipaccess_bts_updown_cb(struct ipa_client_conn *link, int up)
 {
 	struct e1inp_line *line = link->line;
 
-        if (up) {
-                update_fd_settings(line, link->ofd->fd);
-                return;
-        }
+	if (up) {
+		struct osmo_fsm_inst *ka_fsm = ipaccess_line_ts(link->ofd, line)->driver.ipaccess.ka_fsm;
+
+		update_fd_settings(line, link->ofd->fd);
+		if (ka_fsm && line->ipa_kap)
+			ipa_keepalive_fsm_start(ka_fsm);
+		return;
+	}
 
 	if (line->ops->sign_link_down)
 		line->ops->sign_link_down(line);
@@ -701,10 +863,19 @@ int ipaccess_bts_handle_ccm(struct ipa_client_conn *link,
 	struct ipaccess_head *hh = (struct ipaccess_head *) msg->data;
 	struct msgb *rmsg;
 	int ret = 0;
+	/* line might not exist if != bsc||bts */
+	struct e1inp_line *line = link->line;
 
 	/* special handling for IPA CCM. */
 	if (hh->proto == IPAC_PROTO_IPACCESS) {
 		uint8_t msg_type = *(msg->l2h);
+		struct osmo_fsm_inst* ka_fsm = NULL;
+
+		/* peek the pong for our keepalive fsm */
+		if (line && msg_type == IPAC_MSGT_PONG) {
+			ka_fsm = ipaccess_line_ts(link->ofd, line)->driver.ipaccess.ka_fsm;
+			ipa_keepalive_fsm_pong_received(ka_fsm);
+		}
 
 		/* ping, pong and acknowledgment cases. */
 		ret = ipa_ccm_rcvmsg_bts_base(msg, link->ofd);
@@ -830,7 +1001,7 @@ err:
 }
 
 struct ipaccess_line {
-	int line_already_initialized;
+	bool line_already_initialized;
 };
 
 static int ipaccess_line_update(struct e1inp_line *line)
@@ -851,7 +1022,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 	if (il->line_already_initialized)
 		return 0;
 
-	il->line_already_initialized = 1;
+	il->line_already_initialized = true;
 
 	switch(line->ops->cfg.ipa.role) {
 	case E1INP_LINE_R_BSC: {
@@ -895,6 +1066,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 	}
 	case E1INP_LINE_R_BTS: {
 		struct ipa_client_conn *link;
+		struct e1inp_ts *e1i_ts;
 
 		LOGP(DLINP, LOGL_NOTICE, "enabling ipaccess BTS mode, "
 		     "OML connecting to %s:%u\n", line->ops->cfg.ipa.addr,
@@ -922,6 +1094,9 @@ static int ipaccess_line_update(struct e1inp_line *line)
 			ipa_client_conn_destroy(link);
 			return -EIO;
 		}
+
+		e1i_ts = e1inp_line_ipa_oml_ts(line);
+		ipaccess_bts_keepalive_fsm_alloc(e1i_ts, link, "oml_bts_to_bsc");
 		ret = 0;
 		break;
 	}
@@ -944,6 +1119,7 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 				uint8_t trx_nr)
 {
 	struct ipa_client_conn *rsl_link;
+	struct e1inp_ts *e1i_ts = e1inp_line_ipa_rsl_ts(line, trx_nr);
 
 	if (E1INP_SIGN_RSL+trx_nr-1 >= NUM_E1_TS) {
 		LOGP(DLINP, LOGL_ERROR, "cannot create RSL BTS link: "
@@ -972,6 +1148,8 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 		ipa_client_conn_destroy(rsl_link);
 		return -EIO;
 	}
+
+	ipaccess_bts_keepalive_fsm_alloc(e1i_ts, rsl_link, "rsl_bts_to_bsc");
 	return 0;
 }
 
